/*
 * Part of FSView.
 * Example for usage of the treemap widget
 *
 * (C) 2002, Josef Weidendorfer
 */


#include <qpopupmenu.h>
#include <qdir.h>
#include <qtimer.h>

#include <klocale.h>
#include <kconfig.h>
#include <kglobal.h>
#include <kapplication.h>

#include "fsview.h"


// Inode

Inode::Inode(QString p)
{
  init(p);
}

QString Inode::path()
{
  return _info.filePath();
}

void Inode::setPath(QString p)
{
  init(p);
  refresh();
  update();
}

void Inode::init(QString p)
{
  _size = 0;
  _sizeEstimation = 0;
  _subDirsUnknown = 0;
  _files = 0;
  _filesEstimation = 0;

  _info = QFileInfo(p);
  if (_info.isDir()) {
    FSView::getDirMetric(path(), _sizeEstimation, _filesEstimation);
    // we always start with estimations >0
    _sizeEstimation++;
    _filesEstimation++;
  }
  else {
    _size = _info.size();
    _files = 0;
  }
}

unsigned int Inode::value()
{
  if (_sizeEstimation > _size)
    return _sizeEstimation;
  return _size;
}

// this is called after an update(), when the metrics of all
// subdirectories are known
void Inode::finishUpdate()
{
  // create subitem for all real files in the directory...
  QDir d(_info.filePath());
  d.setFilter( QDir::Files | QDir::Hidden | QDir::NoSymLinks );
  QStringList l = d.entryList();
  for(QStringList::Iterator it = l.begin(); it != l.end(); ++it ) {
    Inode* i = new Inode(d.filePath(*it));
    _size += i->value();
    _files++;
    addItem(i);
  }

  TreeMapItem* p = parent();
  if (p && (p!=widget()))
    ((Inode*)p)->gotChildSize(_size, _files);

  int dd = ((FSView*)widget())->pathDepth() + depth();
  if (0)
    qDebug("finishUpdate %s: files %d, depth %d",
           path().ascii(), _files, dd);
  // for "/usr" is dd==3
  if ((_files > 500) ||
      ((_files>50) && (dd<5)) ||
      ((_files>1) && (dd<4)))
    FSView::setDirMetric(path(), _size, _files);

  _sizeEstimation = 0;
  _filesEstimation = 0;
}

void Inode::update()
{
  _size = 0;
  _files = 0;
  _subDirsUnknown = 0;

  if (0) qDebug("Inode::update(%s)", path().ascii());

  // update subdirectories
  QDir d(_info.filePath());
  d.setFilter( QDir::Dirs | QDir::Hidden | QDir::NoSymLinks );
  QStringList l = d.entryList();
  for(QStringList::Iterator it = l.begin(); it != l.end(); ++it ) {
    if (( (*it) == ".") || ( (*it) == "..")) continue;
    Inode* i = new Inode(d.filePath(*it));
    ((FSView*)widget())->requestUpdate(i);
    _subDirsUnknown++;
    _files++; // we count directories, too
    addItem(i);
  }

  if (_subDirsUnknown == 0)  finishUpdate();
}

void Inode::gotChildSize(unsigned int s, unsigned int f)
{
  _size += s;
  _files += f;
  if (_subDirsUnknown > 0) _subDirsUnknown--;

  if (0) qDebug("Inode::gotChildSize(%d,%d) [%s]", s, f, text(0).ascii());
  if (0) qDebug("  new Size %d, new Files %d, unknown SubDirs %d",
                _size, _files, _subDirsUnknown);

  if (_subDirsUnknown == 0) finishUpdate();
}

QColor Inode::color()
{
  int d = ((FSView*)widget())->pathDepth() + depth();
  return QColor((100*d)%360, 128,128, QColor::Hsv);
}

QString Inode::text(int i)
{
  if (i==0) {
    QString name = _info.fileName();
    if (_info.isDir()) name += "/";
    return name;
  }
  if (i==1) {
    QString text;
    unsigned int s = value();

    if (s < 1000)
      text = QString("%1 B").arg(s);
    else if (s < 10000)
      text = QString("%1 kB").arg((double)s/1000+.005,0,'f',2);
    else if (s < 100000)
      text = QString("%1 kB").arg((double)s/1000+.05,0,'f',1);
    else if (s < 1000000)
      text = QString("%1 kB").arg((int)((double)s/1000+.5));
    else if (s < 10000000)
      text = QString("%1 MB").arg((double)s/1000000+.005,0,'f',2);
    else if (s < 100000000)
      text = QString("%1 MB").arg((double)s/1000000+.05,0,'f',1);
    else if (s < 1000000000)
      text = QString("%1 MB").arg((int)((double)s/1000000+.5));
    else
      text =  QString("%1 GB").arg((double)s/1000000000+.005,0,'f',2);

    if (_sizeEstimation>0) text += "+";
    return text;
  }

  if (i==2) {
    QString text;
    unsigned int f = _files;
    if (_filesEstimation > _files) f = _filesEstimation;
    if (f>0) {
      while (f>1000) {
        text = QString("%1 %2").arg(QString::number(f).right(3)).arg(text);
        f /= 1000;
      }
      text = QString("%1 %2").arg(QString::number(f)).arg(text);
      if (_filesEstimation>0) text += "+";
    }
    return text;
  }

  if (i==3) return _info.lastModified().toString();
  if (i==4) return _info.owner();
  if (i==5) return _info.group();
  return QString();
}


// FSView

QMap<QString, QPair<unsigned int, unsigned int> > FSView::_dirMetric;

FSView::FSView()
{
  _pathDepth = 0;
  _allowRefresh = true;

  // restore TreeMap visualization options of last execution
  KConfigGroup tmconfig(KGlobal::config(), QCString("TreeMap"));
  restoreOptions(&tmconfig);

  if (_dirMetric.count() == 0) {
    // restore metric cache
    KConfigGroup cconfig(KGlobal::config(), QCString("MetricCache"));
    int ccount = cconfig.readNumEntry("Count", 0);
    int i, s, f;
    QString str;
    for (i=1;i<=ccount;i++) {
      str = QString("Dir%1").arg(i);
      if (!cconfig.hasKey(str)) continue;
      str = cconfig.readEntry(str);
      s = cconfig.readNumEntry(QString("Size%1").arg(i), 0);
      f = cconfig.readNumEntry(QString("Files%1").arg(i), 0);
      if (s==0 || f==0) continue;
      setDirMetric(str, s, f);
    }
  }
}

void FSView::setPath(QString p)
{
  TreeMapItem* f = childs()->first();
  if (!f) return;

  //qDebug("FSView::setPath %s", p.ascii());

  _path = QFileInfo(p).absFilePath();
  _pathDepth = _path.contains('/');

  ((Inode*)f)->setPath(_path);
  setCaption(QString("%1 - FSView").arg(_path));
}

bool FSView::getDirMetric(QString k, unsigned int& s, unsigned int& f)
{
  QMap<QString, QPair<unsigned int, unsigned int> >::iterator it;

  it = _dirMetric.find(k);
  if (it == _dirMetric.end()) return false;

  s = (*it).first;
  f = (*it).second;

  if (0) qDebug("getDirMetric %s", k.ascii());
  if (0) qDebug(" - got size %d, files %d", s, f);

  return true;
}

void FSView::setDirMetric(QString k, unsigned int s, unsigned int f)
{
  if (0) qDebug("setDirMetric %s: size %d, files %d", k.ascii(), s, f);
  _dirMetric.insert(k, qMakePair(s, f));
}

void FSView::requestUpdate(Inode* i)
{
  if (0) qDebug("FSView::requestUpdate(%s)", i->path().ascii());

  _needsUpdate.append(i);
  if (_needsUpdate.count() == 1)
    QTimer::singleShot(0, this, SLOT(doUpdate()));
}

void FSView::selected(TreeMapItem* i)
{
  setPath(((Inode*)i)->path());
}

void FSView::contextMenu(TreeMapItem* i, const QPoint& p)
{
  QPopupMenu popup;

  // choosing from the selection menu will give a selectionChanged() signal
  popup.insertItem(i18n("Select"),selectionMenu(901,i),900);
  popup.insertItem(i18n("Go up"), 2);
  popup.insertSeparator();
  popup.insertItem(i18n("Stop Refresh"), 3);
  popup.setItemEnabled(3, _needsUpdate.count()>0);
  if (i) popup.insertItem(i18n("Refresh '%1'").arg(i->text(0)), 4);
  popup.insertItem(i18n("Refresh"), 5);
  popup.setItemEnabled(5, _needsUpdate.count()==0);
  popup.insertSeparator();
  popup.insertItem("Stop at Depth",depthStopMenu(1001, i),1000);
  popup.insertItem("Stop at Area",areaStopMenu(1101, i),1100);
  popup.insertItem("Stop at Name",textStopMenu(1201, i),1200);
  popup.insertSeparator();
  popup.insertItem("Options",visualizationMenu(1301),1300);

  _allowRefresh = false;
  int r = popup.exec(mapToGlobal(p));
  _allowRefresh = true;

  if (r==1)
    selected(i);
  else if (r==2) {
    Inode* i = (Inode*) childs()->first();
    if (i) setPath(i->path()+"/..");
  }
  else if (r==3)
    _needsUpdate.clear();
  else if (r==4) {
    ((Inode*)i)->update();
  }
  else if (r==5) {
    Inode* i = (Inode*) childs()->first();
    if (i) i->update();
  }
}

void FSView::quit()
{
  KConfigGroup tmconfig(KGlobal::config(), QCString("TreeMap"));
  saveOptions(&tmconfig);

  KConfigGroup gconfig(KGlobal::config(), QCString("General"));
  gconfig.writeEntry("Path", _path);

  // save metric cache
  KConfigGroup cconfig(KGlobal::config(), QCString("MetricCache"));
  QMap<QString, QPair<unsigned int, unsigned int> >::iterator it;
  int c = 1;
  for (it=_dirMetric.begin();it!=_dirMetric.end();++it) {
    cconfig.writeEntry(QString("Dir%1").arg(c), it.key());
    cconfig.writeEntry(QString("Size%1").arg(c), (*it).first);
    cconfig.writeEntry(QString("Files%1").arg(c), (*it).second);
    c++;
  }
  cconfig.writeEntry("Count", c-1);

  KApplication::kApplication()->quit();
}

void FSView::doUpdate()
{
  static int redrawCounter = 0;
  int inodeCounter = 0;
  Inode* i = 0;

  while (inodeCounter < 10) {
    i = _needsUpdate.first();
    if (!i) break;

    if (0) qDebug("doUpdate: %s", i->path().ascii());
    i->update();

    _needsUpdate.removeFirst();
    inodeCounter++;
  }

  if (_allowRefresh) {
    redrawCounter++;
    if (redrawCounter>20 || !i) {
      redrawCounter = 0;

      if (0) qDebug("doUpdate: Redraw");
      redraw();
    }
  }

  if (_needsUpdate.count()>0)
    QTimer::singleShot(0, this, SLOT(doUpdate()));
}
