/**
 * A Widget for visualizing hierarchical metrics as areas.
 * The API is similar to QListView.
 *
 * JW, 2002
 */

#include <math.h>

#include <qpainter.h>
#include <qtooltip.h>
#include <qregexp.h>
#include <qstyle.h>
#include <qpopupmenu.h>

#include <klocale.h>
#include <kconfig.h>

#include "treemap.h"


#define MAX_TREEMAP_TEXT 12

int TreeMapItemList::compareItems ( Item item1, Item item2 )
{
  bool ascending;
  int result;

  int textNo = ((TreeMapItem*)item1)->sorting(&ascending);

  if (textNo < 0)
    result = ((TreeMapItem*)item1)->value() -
             ((TreeMapItem*)item2)->value();
  else
    result = ((TreeMapItem*)item1)->text(textNo) <
             ((TreeMapItem*)item2)->text(textNo);

  return ascending ? result : -result;
}



// TreeMapItem

TreeMapItem::TreeMapItem(TreeMapItem* parent, unsigned int value)
{
  _value = value;
  _parent = parent;

  _sum = 0;
  _childs = 0;
  _widget = 0;
  _index = -1;
  _depth = -1; // not set
  _unused_self = 0;

  if (_parent)
    addItem(_parent);
}

TreeMapItem::TreeMapItem(TreeMapItem* parent, unsigned int value,
                         QString text1, QString text2,
                         QString text3, QString text4)
{
  _value = value;
  _parent = parent;

  // this resizes the text vector only if needed
  if (!text4.isEmpty()) setText(3, text4);
  if (!text3.isEmpty()) setText(2, text3);
  if (!text2.isEmpty()) setText(1, text2);
  setText(0, text1);

  _sum = 0;
  _childs = 0;
  _widget = 0;
  _index = -1;
  _depth = -1; // not set
  _unused_self = 0;

  if (_parent) addItem(_parent);
}

TreeMapItem::~TreeMapItem()
{
  // notify widget about deletion
  if (_widget) _widget->deletingItem(this);

  if (_childs) delete _childs;
}

void TreeMapItem::setParent(TreeMapItem* p)
{
  _parent = p;
  if (p) _widget = p->_widget;
}

bool TreeMapItem::isChildOf(TreeMapItem* item)
{
  if (!item) return false;

  TreeMapItem* i = this;
  while (i) {
    if (item == i) return true;
    i = i->_parent;
  }
  return false;
}

TreeMapItem* TreeMapItem::commonParent(TreeMapItem* item)
{
  while (item && !isChildOf(item)) {
    item = item->parent();
  }
  return item;
}

void TreeMapItem::redraw()
{
  if (_widget)
    _widget->redraw(this);
}

// invalidates current childs and forces redraw
// this is only usefull when childs are created on demand in items()
void TreeMapItem::refresh()
{
  if (_childs) {
    delete _childs;
    _childs = 0;
  }
  redraw();
}


QStringList TreeMapItem::path(int textNo)
{
  QStringList list(text(textNo));

  TreeMapItem* i = _parent;
  while (i) {
    QString text = i->text(textNo);
    if (!text.isEmpty())
        list.prepend(i->text(textNo));
    i = i->_parent;
  }
  return list;
}

int TreeMapItem::depth()
{
  if (_depth>0) return _depth;

  if (_parent)
    return _parent->depth() + 1;
  return 1;
}


bool TreeMapItem::initialized()
{
  if (!_childs) {
    _childs = new TreeMapItemList;
    _childs->setAutoDelete(true);
    return false;
  }
  return true;
}

void TreeMapItem::addItem(TreeMapItem* i)
{
  if (!i) return;

  if (!_childs) {
    _childs = new TreeMapItemList;
    _childs->setAutoDelete(true);
  }
  i->setParent(this);

  if (sorting(0) == -1)
    _childs->append(i); // preserve insertion order
  else
    _childs->inSort(i);
}

void TreeMapItem::setText(int textNo, QString text)
{
  if (textNo>=0 && textNo<MAX_TREEMAP_TEXT) {
    if ((int)_text.size() < textNo+1)
      _text.resize(textNo+1, QString::null);
    _text[textNo] = text;
  }
}

void TreeMapItem::setPixmap(int textNo, QPixmap pix)
{
  if (textNo>=0 && textNo<MAX_TREEMAP_TEXT) {
    if ((int)_pix.size() < textNo+1)
      _pix.resize(textNo+1, QPixmap());
    _pix[textNo] = pix;
  }
}

// default implementations of virtual functions

unsigned int TreeMapItem::value()
{
  return _value;
}

unsigned int TreeMapItem::sum()
{
  return _sum;
}

QString TreeMapItem::text(int textNo)
{
  if (textNo<0 || (int)_text.size()<textNo+1)
    return QString::null;

  return _text[textNo];
}

QPixmap TreeMapItem::pixmap(int textNo)
{
  if (textNo<0 || (int)_pix.size()<textNo+1)
    return QPixmap();

  return _pix[textNo];
}

int TreeMapItem::borderWidth()
{
  if (_widget)
    return _widget->borderWidth();

  return 2;
}

int TreeMapItem::sorting(bool* ascending)
{
  if (_widget)
    return _widget->sorting(ascending);

  if (ascending) *ascending = true;
  return -2;
}

TreeMapItem::SplitMode TreeMapItem::splitMode()
{
  if (_widget)
    return _widget->splitMode();

  return Best;
}

int TreeMapItem::rtti()
{
  return 0;
}

TreeMapItemList* TreeMapItem::childs()
{
  if (!_childs) {
    _childs = new TreeMapItemList;
    _childs->setAutoDelete(true);
  }
  return _childs;
}

QColor TreeMapItem::color()
{
  return _color;
}

TreeMapItem::Position TreeMapItem::position(int textNo)
{
  if (_widget)
    return _widget->position(textNo);

  switch(textNo%4) {
  case 0: return TopLeft;
  case 1: return TopRight;
  case 2: return BottomRight;
  case 3: return BottomLeft;
  default:break;
  }
  return TopLeft;
}

// Tooltips for TreeMapWidget

class TreeMapTip: public QToolTip
{
public:
  TreeMapTip( QWidget* p ):QToolTip(p) {}

protected:
    void maybeTip( const QPoint & );
};

void TreeMapTip::maybeTip( const QPoint& pos )
{
  if ( !parentWidget()->inherits( "TreeMapWidget" ) )
        return;

  TreeMapWidget* p = (TreeMapWidget*)parentWidget();
  TreeMapItem* i;
  i = p->item(pos.x(), pos.y());
  if (i) {
    if (i->freeRect1().contains(pos))
      tip(i->freeRect1(), p->tipString(i));
    else if (i->freeRect2().contains(pos))
      tip(i->freeRect2(), p->tipString(i));
    else if (i->freeRect3().contains(pos))
      tip(i->freeRect3(), p->tipString(i));
  }
}



// TreeMapWidget

TreeMapWidget::TreeMapWidget(QWidget* parent, const char* name)
  : QWidget(parent, name), TreeMapItem()
{
  setWidget(this);

  // default behaviour
  _selectionMode = Single;
  _splitMode = AlwaysBest;
  _visibleWidth = 2;
  _reuseSpace = false;
  _skipIncorrectBorder = false;
  _drawSeparators = false;
  _sortAscending = false;
  _sortTextNo = -2; // sort by value()
  _allowRotation = true;
  _borderWidth = 2;
  _shading = true; // beautiful is default!
  _maxSelectDepth = -1; // unlimited
  _maxDrawingDepth = -1; // unlimited
  _minimalArea = -1; // unlimited

  // _stopAtText will be unset on resizing (per default)
  // _textVisible will be true on resizing (per default)
  // _forceText will be false on resizing (per default)

  // start state: _selection is an empty list
  _current = 0;
  _oldCurrent = 0;
  _pressed = 0;
  _lastOver = 0;
  _needsRefresh = this;

  setBackgroundMode(Qt::NoBackground);
  setFocusPolicy(QWidget::StrongFocus);
  _tip = new TreeMapTip(this);
}

TreeMapWidget::~TreeMapWidget()
{
}

void TreeMapWidget::setSplitMode(SplitMode m)
{
  if (_splitMode == m) return;

  _splitMode = m;
  redraw();
}

TreeMapItem::SplitMode TreeMapWidget::splitMode()
{
  return _splitMode;
}

bool TreeMapWidget::setSplitMode(QString mode)
{
  if (mode == "Bisection") setSplitMode(Bisection);
  else if (mode == "Columns") setSplitMode(Columns);
  else if (mode == "Rows") setSplitMode(Rows);
  else if (mode == "AlwaysBest") setSplitMode(AlwaysBest);
  else if (mode == "Best")       setSplitMode(Best);
  else if (mode == "HAlternate") setSplitMode(HAlternate);
  else if (mode == "VAlternate") setSplitMode(VAlternate);
  else if (mode == "Horizontal") setSplitMode(TreeMapItem::Horizontal);
  else if (mode == "Vertical")   setSplitMode(TreeMapItem::Vertical);
  else return false;

  return true;
}

QString TreeMapWidget::splitModeString()
{
  QString mode;
  switch(splitMode()) {
    case Bisection:  mode = "Bisection"; break;
    case Columns:    mode = "Columns"; break;
    case Rows:       mode = "Rows"; break;
    case AlwaysBest: mode = "AlwaysBest"; break;
    case Best:       mode = "Best"; break;
    case HAlternate: mode = "HAlternate"; break;
    case VAlternate: mode = "VAlternate"; break;
    case TreeMapItem::Horizontal: mode = "Horizontal"; break;
    case TreeMapItem::Vertical:   mode = "Vertical"; break;
    default: mode = "Unknown"; break;
  }
  return mode;
}


void TreeMapWidget::setShadingEnabled(bool s)
{
  if (_shading == s) return;

  _shading = s;
  redraw();
}

void TreeMapWidget::setAllowRotation(bool enable)
{
  if (_allowRotation == enable) return;

  _allowRotation = enable;
  redraw();
}

void TreeMapWidget::setVisibleWidth(int width, bool reuseSpace)
{
  if (_visibleWidth == width && _reuseSpace == reuseSpace) return;

  _visibleWidth = width;
  _reuseSpace = reuseSpace;
  redraw();
}

void TreeMapWidget::setSkipIncorrectBorder(bool enable)
{
  if (_skipIncorrectBorder == enable) return;

  _skipIncorrectBorder = enable;
  redraw();
}

void TreeMapWidget::setMaxDrawingDepth(int d)
{
  if (_maxDrawingDepth == d) return;

  _maxDrawingDepth = d;
  redraw();
}

QString TreeMapWidget::defaultTextType(int textNo)
{
  return i18n("Text %1").arg(textNo+1);
}

QString TreeMapWidget::defaultTextStop(int)
{
  return QString();
}

bool TreeMapWidget::defaultTextVisible(int textNo)
{
  return (textNo<2);
}

bool TreeMapWidget::defaultTextForced(int)
{
  return false;
}

TreeMapItem::Position TreeMapWidget::defaultPosition(int textNo)
{
  switch(textNo%4) {
  case 0: return TopLeft;
  case 1: return TopRight;
  case 2: return BottomRight;
  case 3: return BottomLeft;
  default:break;
  }
  return TopLeft;
}

bool TreeMapWidget::resizeAttr(int size)
{
  if (size<0 || size>=MAX_TREEMAP_TEXT) return false;

  if (size>(int)_attr.size()) {
    struct TextAttr a;
    int oldSize = _attr.size();
    _attr.resize(size, a);
    while (oldSize<size) {
      _attr[oldSize].type    = defaultTextType(oldSize);
      _attr[oldSize].stop    = defaultTextStop(oldSize);
      _attr[oldSize].visible = defaultTextVisible(oldSize);
      _attr[oldSize].forced  = defaultTextForced(oldSize);
      _attr[oldSize].pos     = defaultPosition(oldSize);
      oldSize++;
    }
  }
  return true;
}

void TreeMapWidget::setTextType(int textNo, QString type)
{
  if (((int)_attr.size() < textNo+1) &&
      (type == defaultTextType(textNo))) return;
  if (resizeAttr(textNo+1)) _attr[textNo].type = type;

  // no need to redraw: the type string is not visible in the TreeMap
}

QString TreeMapWidget::textType(int textNo)
{
  if (textNo<0 || (int)_attr.size()<textNo+1) return defaultTextType(textNo);
  return _attr[textNo].type;
}

void TreeMapWidget::setTextStop(int textNo, QString stop)
{
  if (((int)_attr.size() < textNo+1) &&
      (stop == defaultTextStop(textNo))) return;
  if (resizeAttr(textNo+1)) {
    _attr[textNo].stop = stop;
    redraw();
  }
}

QString TreeMapWidget::textStop(int textNo)
{
  if (textNo<0 || (int)_attr.size()<textNo+1) return defaultTextStop(textNo);
  return _attr[textNo].stop;
}

void TreeMapWidget::setTextVisible(int textNo, bool enable)
{
  if (((int)_attr.size() < textNo+1) &&
      (enable == defaultTextVisible(textNo))) return;

  if (resizeAttr(textNo+1)) {
    _attr[textNo].visible = enable;
    redraw();
  }
}

bool TreeMapWidget::textVisible(int textNo)
{
  if (textNo<0 || (int)_attr.size()<textNo+1)
    return defaultTextVisible(textNo);

  return _attr[textNo].visible;
}

void TreeMapWidget::setTextForced(int textNo, bool enable)
{
  if (((int)_attr.size() < textNo+1) &&
      (enable == defaultTextForced(textNo))) return;

  if (resizeAttr(textNo+1)) {
    _attr[textNo].forced = enable;
    if (_attr[textNo].visible) redraw();
  }
}

bool TreeMapWidget::textForced(int textNo)
{
  if (textNo<0 || (int)_attr.size()<textNo+1)
    return defaultTextForced(textNo);

  return _attr[textNo].forced;
}

void TreeMapWidget::setPosition(int textNo, TreeMapItem::Position pos)
{
  if (((int)_attr.size() < textNo+1) &&
      (pos == defaultPosition(textNo))) return;

  if (resizeAttr(textNo+1)) {
    _attr[textNo].pos = pos;
    if (_attr[textNo].visible) redraw();
  }
}

TreeMapItem::Position TreeMapWidget::position(int textNo)
{
  if (textNo<0 || (int)_attr.size()<textNo+1)
    return defaultPosition(textNo);

  return _attr[textNo].pos;
}

void TreeMapWidget::setPosition(int textNo, QString pos)
{
  if (pos == "TopLeft") setPosition(textNo, TreeMapItem::TopLeft);
  else if (pos == "TopRight") setPosition(textNo, TreeMapItem::TopRight);
  else if (pos == "BottomLeft") setPosition(textNo, TreeMapItem::BottomLeft);
  else if (pos == "BottomRight") setPosition(textNo, TreeMapItem::BottomRight);
}

QString TreeMapWidget::positionString(int textNo)
{
  TreeMapItem::Position pos = position(textNo);
  if (pos == TreeMapItem::TopLeft) return QString("TopLeft");
  if (pos == TreeMapItem::TopRight) return QString("TopRight");
  if (pos == TreeMapItem::BottomLeft) return QString("BottomLeft");
  if (pos == TreeMapItem::BottomRight) return QString("BottomRight");
  return QString("unknown");
}

void TreeMapWidget::setMinimalArea(int area)
{
  if (_minimalArea == area) return;

  _minimalArea = area;
  redraw();
}


// to be called directly after constructor
void TreeMapWidget::setSorting(int textNo, bool ascending)
{
  _sortAscending = ascending;
  _sortTextNo = textNo;
}

int TreeMapWidget::sorting(bool* ascending)
{
  if (ascending) *ascending = _sortAscending;
  return _sortTextNo;
}

void TreeMapWidget::deletingItem(TreeMapItem* i)
{
  // remove any references to the item to be deleted
  while(_selection.findRef(i) > -1)
    _selection.remove();

  while(_tmpSelection.findRef(i) > -1)
    _tmpSelection.remove();

  if (_current == i) _current = 0;
  if (_oldCurrent == i) _oldCurrent = 0;
  if (_pressed == i) _pressed = 0;
  if (_lastOver == i) _lastOver = 0;
}


QString TreeMapWidget::tipString(TreeMapItem* i)
{
  QString tip, itemTip;

  while (i) {
    if (!i->text(0).isEmpty()) {
      itemTip = i->text(0);
      if (!i->text(1).isEmpty())
        itemTip += " (" + i->text(1) + ")";

      if (!tip.isEmpty())
        tip += "\n";

      tip += itemTip;
    }
    i = i->parent();
  }
  return tip;
}

TreeMapItem* TreeMapWidget::item(int x, int y)
{
  TreeMapItem* p = this;
  TreeMapItem* i;

  if (!rect().contains(x, y)) return 0;
  while (1) {
    TreeMapItemList* list = p->childs();
    if (!list)
      i = 0;
    else {
      int idx=0;
      for (i=list->first();i;i=list->next(),idx++) {
        if (i->itemRect().contains(x, y)) {
          p->setIndex(idx);
          break;
        }
      }
    }

    if (!i) {
      static TreeMapItem* last = 0;
      if (p != last) {
        last = p;
        /*
        qDebug("item(%d,%d): Got %s (Size %dx%d, Val %d)",
               x, y, p->path().ascii(),
               p->itemRect().width(), p->itemRect().height(), p->value());
        */
      }

      return p;
    }
    p = i;
  }
  return 0;
}

TreeMapItem* TreeMapWidget::possibleSelection(TreeMapItem* i)
{
  if (i) {
    if (_maxSelectDepth>=0) {
      int depth = i->depth();
      while(i && depth > _maxSelectDepth) {
        i = i->parent();
        depth--;
      }
    }
  }
  return i;
}

TreeMapItem* TreeMapWidget::visibleItem(TreeMapItem* i)
{
  if (i) {
    /* Must have a visible area */
    while(i && ((i->itemRect().width() <1) ||
                (i->itemRect().height() <1))) {
      TreeMapItem* p = i->parent();
      if (!p) break;
      int idx = p->childs()->findRef(i);
      idx--;
      if (idx<0)
        i = p;
      else
        i = p->childs()->at(idx);
    }
  }
  return i;
}

void TreeMapWidget::setSelected(TreeMapItem* item, bool selected)
{
  setCurrent(item);
  item = possibleSelection(item);
  if (setTmpSelected(item, selected)) {
    _selection = _tmpSelection;
    if (_selectionMode == Single)
      emit selectionChanged(item);
    emit selectionChanged();
    redraw();

    /*
    qDebug("%selected Item %s (depth %d)",
           selected ? "S":"Des",
           item ? item->path().ascii() : "(null)",
           item ? item->depth() : -1);
    */
  }
}

// only modifies _tmpSelection
// returns true on a change
bool TreeMapWidget::setTmpSelected(TreeMapItem* item, bool selected)
{
  if (_selectionMode == NoSelection) return false;

  bool wasSelected = isTmpSelected(item);

  if (_selectionMode == Single) {
    int oldCount = _tmpSelection.count();
    _tmpSelection.clear();
    if (selected)
      _tmpSelection.append(item);
    return !(wasSelected == selected) || (oldCount>1);
  }

  if (selected) {
    TreeMapItem* i=_tmpSelection.first();
    while (i) {
      if (i->isChildOf(item) || item->isChildOf(i)) {
        _tmpSelection.remove();
        i = _tmpSelection.current();
      }
      else
        i = _tmpSelection.next();
    }
    _tmpSelection.append(item);
  }
  else
    _tmpSelection.removeRef(item);

  return !(wasSelected == selected);
}


void TreeMapWidget::clearSelection(TreeMapItem* parent)
{
  bool changed = false;
  TreeMapItem* i=_selection.first();
  while (i) {
    if (i->isChildOf(parent)) {
      _selection.remove();
      changed = true;
      i = _selection.current();
    }
    else
      i = _selection.next();
  }
  if (changed)
    emit selectionChanged();
}

bool TreeMapWidget::isSelected(TreeMapItem* i)
{
  return _selection.containsRef(i)>0;
}

bool TreeMapWidget::isTmpSelected(TreeMapItem* i)
{
  return _tmpSelection.containsRef(i)>0;
}


void TreeMapWidget::setCurrent(TreeMapItem* i, bool kbd)
{
  if (i == _current) return;

  TreeMapItem* old = _current;
  _current = i;

  if (old) old->redraw();
  if (i) i->redraw();

  //qDebug("Current Item %s", i ? i->path().ascii() : "(null)");

  emit currentChanged(i, kbd);
}

void TreeMapWidget::setRangeSelection(TreeMapItem* i1,
                                         TreeMapItem* i2, bool selected)
{
  setCurrent(i2);

  i1 = possibleSelection(i1);
  i2 = possibleSelection(i2);
  if (setTmpRangeSelection(i1, i2, selected)) {
    _selection = _tmpSelection;
    if (_selectionMode == Single)
      emit selectionChanged(i2);
    emit selectionChanged();
    redraw();
  }
}

bool TreeMapWidget::setTmpRangeSelection(TreeMapItem* i1,
                                            TreeMapItem* i2,
                                            bool selected)
{
  if (i1->isChildOf(i2)) return setTmpSelected(i2, selected);
  if (i2->isChildOf(i1)) return setTmpSelected(i1, selected);

  bool changed = false;
  if (setTmpSelected(i1, selected)) changed = true;
  if (setTmpSelected(i2, selected)) changed = true;

  TreeMapItem* commonParent = i1;
  while (commonParent && !i2->isChildOf(commonParent)) {
    i1 = commonParent;
    commonParent = commonParent->parent();
  }
  if (!commonParent) return changed;
  while (i2 && i2->parent() != commonParent)
    i2 = i2->parent();
  if (!i2) return changed;

  TreeMapItemList* list = commonParent->childs();
  if (list) {
    TreeMapItem* i = list->first();
    bool between = false;
    while (i) {
      if (between) {
        if (i==i1 || i==i2) break;
        if (setTmpSelected(i, selected)) changed = true;
      }
      else if (i==i1 || i==i2)
        between = true;
      i = list->next();
    }
  }
  return changed;
}

void TreeMapWidget::contextMenuEvent( QContextMenuEvent* e )
{
  //qDebug("TreeMapWidget::contextMenuEvent");

  if ( receivers( SIGNAL(contextMenuRequested(TreeMapItem*, const QPoint &)) ) )
    e->accept();

  if ( e->reason() == QContextMenuEvent::Keyboard ) {
    QRect r = (_current) ? _current->itemRect() : itemRect();
    QPoint p = QPoint(r.left() + r.width()/2, r.top() + r.height()/2);
    emit contextMenuRequested(_current, p);
  }
  else {
    TreeMapItem* i = item(e->x(), e->y());
    emit contextMenuRequested(i, e->pos());
  }
}


void TreeMapWidget::mousePressEvent( QMouseEvent* e )
{
  //qDebug("TreeMapWidget::mousePressEvent");

  _oldCurrent = _current;

  TreeMapItem* i = item(e->x(), e->y());
  if (e->button() == RightButton) {
    setCurrent(i);
    emit rightButtonPressed(i, e->pos() );
  }
  if (e->button() != LeftButton) return;

  _pressed = i;

  _inShiftDrag = e->state() & ShiftButton;
  _inControlDrag = e->state() & ControlButton;
  _lastOver = _pressed;

  bool changed = false;
  TreeMapItem* item = possibleSelection(_pressed);

  switch(_selectionMode) {
  case Single:
    changed = setTmpSelected(item, true);
    break;
  case Multi:
    changed = setTmpSelected(item, !isTmpSelected(item));
    break;
  case Extended:
    if (_inControlDrag)
      changed = setTmpSelected(item, !isTmpSelected(item));
    else if (_inShiftDrag) {
      TreeMapItem* sCurrent = possibleSelection(_current);
      changed = setTmpRangeSelection(sCurrent, item,
                                     !isTmpSelected(item));
    }
    else {
      _selectionMode = Single;
      changed = setTmpSelected(item, true);
      _selectionMode = Extended;
    }
    break;
  default:
    break;
  }

  setCurrent(_pressed);

  if (changed)
    redraw();
}

void TreeMapWidget::mouseMoveEvent( QMouseEvent* e )
{
  //qDebug("TreeMapWidget::mouseMoveEvent");

  if (!_pressed) return;
  TreeMapItem* over = item(e->x(), e->y());
  if (_lastOver == over) return;

  setCurrent(over);
  if (over == 0) {
    _lastOver = 0;
    return;
  }

  bool changed = false;
  TreeMapItem* item = possibleSelection(over);

  switch(_selectionMode) {
  case Single:
    changed = setTmpSelected(item, true);
    break;
  case Multi:
    changed = setTmpSelected(item, !isTmpSelected(item));
    break;
  case Extended:
    if (_inControlDrag)
      changed = setTmpSelected(item, !isTmpSelected(item));
    else {
      TreeMapItem* sLast = possibleSelection(_lastOver);
      changed = setTmpRangeSelection(sLast, item, true);
    }
    break;

  default:
    break;
  }

  _lastOver = over;

  if (changed)
    redraw();
}

void TreeMapWidget::mouseReleaseEvent( QMouseEvent* )
{
  //qDebug("TreeMapWidget::mouseReleaseEvent");

  if (!_pressed) return;

  if (!_lastOver) {
    // take back
    setCurrent(_oldCurrent);
    _tmpSelection = _selection;
    redraw();
  }
  else {
    if (! (_tmpSelection == _selection)) {
      _selection = _tmpSelection;
      if (_selectionMode == Single)
        emit selectionChanged(_lastOver);
      emit selectionChanged();
    }
  }

  _pressed = 0;
  _lastOver = 0;
}


void TreeMapWidget::mouseDoubleClickEvent( QMouseEvent* e )
{
  TreeMapItem* over = item(e->x(), e->y());

  emit doubleClicked(over);
}


/* returns -1 if nothing visible found */
int nextVisible(TreeMapItem* i)
{
  TreeMapItem* p = i->parent();
  if (!p || p->itemRect().isEmpty()) return -1;

  int idx = p->childs()->findRef(i);
  if (idx<0) return -1;

  while (idx < (int)p->childs()->count()-1) {
    idx++;
    QRect r = p->childs()->at(idx)->itemRect();
    if (r.width()>1 && r.height()>1)
      return idx;
  }
  return -1;
}

/* returns -1 if nothing visible found */
int prevVisible(TreeMapItem* i)
{
  TreeMapItem* p = i->parent();
  if (!p || p->itemRect().isEmpty()) return -1;

  int idx = p->childs()->findRef(i);
  if (idx<0) return -1;

  while (idx > 0) {
    idx--;
    QRect r = p->childs()->at(idx)->itemRect();
    if (r.width()>1 && r.height()>1)
      return idx;
  }
  return -1;
}




void TreeMapWidget::keyPressEvent( QKeyEvent* e )
{
  if (e->key() == Key_Escape && _pressed) {

    // take back
    if (_oldCurrent != _lastOver)
      setCurrent(_oldCurrent);
    if (! (_tmpSelection == _selection)) {
      _tmpSelection = _selection;
      redraw();
    }
    _pressed = 0;
    _lastOver = 0;
  }

  if (e->key() == Key_Space) {

    switch(_selectionMode) {
    case NoSelection:
      break;
    case Single:
      setSelected(_current, true);
      break;
    case Multi:
      setSelected(_current, !isSelected(_current));
      break;
    case Extended:
      if ((e->state() & ControlButton) || (e->state() & ShiftButton))
        setSelected(_current, !isSelected(_current));
      else {
        _selectionMode = Single;
        setSelected(_current, true);
        _selectionMode = Extended;
      }
    }
    return;
  }

  if (!_current) {
    if (e->key() == Key_Return) {
      setCurrent(this, true);
    }
    return;
  }

  TreeMapItem* old = _current,  *newItem;
  TreeMapItem* p = _current->parent();

  bool goBack;
  if (_current->sorting(&goBack) == -1) {
    // noSorting
    goBack = false;
  }


  if (e->key() == Key_Backspace) {
    newItem = visibleItem(p);
    setCurrent(newItem, true);
  }
  else if (e->key() == Key_Left || e->key() == Key_Down) {
    int newIdx = goBack ? nextVisible(_current) : prevVisible(_current);
    if (p && newIdx>=0) {
      p->setIndex(newIdx);
      setCurrent(p->childs()->at(newIdx), true);
    }
  }
  else if (e->key() == Key_Right || e->key() == Key_Up) {
    int newIdx = goBack ? prevVisible(_current) : nextVisible(_current);
    if (p && newIdx>=0) {
      p->setIndex(newIdx);
      setCurrent(p->childs()->at(newIdx), true);
    }
  }
  else if (e->key() == Key_Return) {
    if (_current->childs() && _current->childs()->count()>0) {
      int newIdx = _current->index();
      if (newIdx<0)
        newIdx = goBack ? (_current->childs()->count()-1) : 0;
      if (newIdx>=(int)_current->childs()->count())
        newIdx = _current->childs()->count()-1;
      newItem = visibleItem(_current->childs()->at(newIdx));
      setCurrent(newItem, true);
    }
  }

  if (old == _current) return;
  if (! (e->state() & ControlButton)) return;
  if (! (e->state() & ShiftButton)) return;

  switch(_selectionMode) {
  case NoSelection:
    break;
  case Single:
    setSelected(_current, true);
    break;
  case Multi:
    setSelected(_current, !isSelected(_current));
    break;
  case Extended:
    if (e->state() & ControlButton)
      setSelected(_current, !isSelected(_current));
    else
      setSelected(_current, isSelected(old));
  }
}

void TreeMapWidget::resizeEvent( QResizeEvent * )
{
  // this automatically redraws (as size is changed)
  drawTreeMap();
}

void TreeMapWidget::paintEvent( QPaintEvent * )
{
  drawTreeMap();
}

void TreeMapWidget::showEvent( QShowEvent * )
{
  // refresh only if needed
  drawTreeMap();
}

// Updates screen from shadow buffer,
// but redraws before if needed
void TreeMapWidget::drawTreeMap()
{
  // no need to draw if hidden
  if (!isVisible()) return;

  if (_pixmap.size() != size())
    _needsRefresh = this;

  if (_needsRefresh) {
    if (_needsRefresh == this) {
      // redraw whole widget
      _pixmap = QPixmap(size());
      _pixmap.fill(backgroundColor());
    }
    QPainter p(&_pixmap);
    if (_needsRefresh == this) {
      p.setPen(black);
      p.drawRect(QRect(2, 2, QWidget::width()-4, QWidget::height()-4));
      setItemRect(QRect(3, 3, QWidget::width()-6, QWidget::height()-6));
    }
    else {
      // only subitem
      if (!_needsRefresh->itemRect().isValid()) return;
    }
    drawItems(&p, _needsRefresh);
    _needsRefresh = 0;
  }

  bitBlt( this, 0, 0, &_pixmap, 0, 0,
          QWidget::width(), QWidget::height(), CopyROP, true);

  if (hasFocus()) {
    QPainter p(this);
    style().drawPrimitive( QStyle::PE_FocusRect, &p,
                           QRect(0, 0, QWidget::width(), QWidget::height()),
                           colorGroup() );
  }
}



void TreeMapWidget::redraw(TreeMapItem* i)
{
  if (!i) return;

  if (!_needsRefresh || (i==this))
    _needsRefresh = i;
  else {
    if (!i->isChildOf(_needsRefresh))
      _needsRefresh = _needsRefresh->commonParent(i);
  }

  if (isVisible()) {
    // delayed drawing if we have multiple redraw requests
    update();
  }
}

void TreeMapWidget::drawItem(QPainter* p,
                             TreeMapItem* item)
{
  QRect r = item->itemRect();
  if (r.width()<=0 || r.height()<=0) return;

  QColor normal = item->color();

  if (item != this) {
    TreeMapItem* i;
    for (i=_tmpSelection.first();i;i=_tmpSelection.next()) {
      if (i->isChildOf(item) || item->isChildOf(i)) {
        normal = normal.light();
        break;
      }
    }
  }

  bool isCurrent = _current && item->isChildOf(_current);

  // 3D raised/sunken frame effect...
  QColor high = normal.light();
  QColor low = normal.dark();
  p->setPen( isCurrent ? low:high);
  p->drawLine(r.left(), r.top(), r.right(), r.top());
  p->drawLine(r.left(), r.top(), r.left(), r.bottom());
  p->setPen( isCurrent ? high:low);
  p->drawLine(r.right(), r.top(), r.right(), r.bottom());
  p->drawLine(r.left(), r.bottom(), r.right(), r.bottom());
  r.setRect(r.x()+1, r.y()+1, r.width()-2, r.height()-2);
  if (r.width()<=0 || r.height()<=0) return;

#if 0
  // red frame when selected
  if (_selected && _selected->isChildOf(item) ) {
    p->setPen(red);
    p->setBrush(NoBrush);
    p->drawRect(r);
    r.setRect(r.x()+1, r.y()+1, r.width()-2, r.height()-2);
    if (r.width()<=0 || r.height()<=0) return;
  }
#endif

  if (_shading) {
    // some shading
    bool goDark = qGray(normal.rgb())>128;
    int rBase, gBase, bBase;
    normal.rgb(&rBase, &gBase, &bBase);
    p->setBrush(NoBrush);

    // shade parameters:
    int d = 7;
    float factor = 0.1, forth=0.7, back1 =0.9, toBack2 = .7, back2 = 0.97;

    // coefficient corrections because of rectangle size
    int s = r.width();
    if (s > r.height()) s = r.height();
    if (s<100) {
      forth -= .3  * (100-s)/100;
      back1 -= .2  * (100-s)/100;
      back2 -= .02 * (100-s)/100;
    }


    // maximal color difference
    int rDiff = goDark ? -rBase/d : (255-rBase)/d;
    int gDiff = goDark ? -gBase/d : (255-gBase)/d;
    int bDiff = goDark ? -bBase/d : (255-bBase)/d;

    QColor shadeColor;
    while (factor<.95) {
      shadeColor.setRgb((int)(rBase+factor*rDiff+.5),
                        (int)(gBase+factor*gDiff+.5),
                        (int)(bBase+factor*bDiff+.5));
      p->setPen(shadeColor);
      p->drawRect(r);
      r.setRect(r.x()+1, r.y()+1, r.width()-2, r.height()-2);
      if (r.width()<=0 || r.height()<=0) return;
      factor = 1.0 - ((1.0 - factor) * forth);
    }

    // and back (1st half)
    while (factor>toBack2) {
      shadeColor.setRgb((int)(rBase+factor*rDiff+.5),
                        (int)(gBase+factor*gDiff+.5),
                        (int)(bBase+factor*bDiff+.5));
      p->setPen(shadeColor);
      p->drawRect(r);
      r.setRect(r.x()+1, r.y()+1, r.width()-2, r.height()-2);
      if (r.width()<=0 || r.height()<=0) return;
      factor = 1.0 - ((1.0 - factor) / back1);
    }

    // and back (2nd half)
    while ( factor>.01) {
      shadeColor.setRgb((int)(rBase+factor*rDiff+.5),
                        (int)(gBase+factor*gDiff+.5),
                        (int)(bBase+factor*bDiff+.5));
      p->setPen(shadeColor);
      p->drawRect(r);
      r.setRect(r.x()+1, r.y()+1, r.width()-2, r.height()-2);
      if (r.width()<=0 || r.height()<=0) return;

      factor = factor * back2;
    }
  }

  // fill inside
  p->setPen(NoPen);
  p->setBrush(normal);
  p->drawRect(r);
}

bool TreeMapWidget::horizontal(TreeMapItem* i, const QRect& r)
{
  switch(i->splitMode()) {
  case HAlternate:
    return (i->depth()%2)==1;
  case VAlternate:
    return (i->depth()%2)==0;
  case TreeMapItem::Horizontal:
    return true;
  case TreeMapItem::Vertical:
    return false;
  default:
    return r.width() > r.height();
  }
  return false;
}

// return unused space on last line.
// this can be used when this function is called with reversed aligning
// by setting lastUnused to this value.
int TreeMapWidget::drawString(QPainter* p, TreeMapItem* item, int textNo,
                              QRect& r, int unused, bool right, bool bottom)
{
  if (item->text(textNo).isEmpty()) return 0;

  int h = fontMetrics().height();

  bool rotate = (_allowRotation && (r.height() > r.width()));
  int width   = (rotate ? r.height() : r.width()) -4;
  int height  = (rotate ? r.width() : r.height());
  int lines   = height / h;
  QString remaining, name = item->text(textNo);
  int y = bottom ? height - h : 0;
  int w = 0;

  static int dotW = 0;
  if (!dotW) dotW = fontMetrics().width("...");
  if (width < dotW) return 0;

  QPixmap pix = item->pixmap(textNo);
  int pixW = pix.width();
  bool pixDrawn = true;
  if (pixW>0) {
    pixW += 5; // distance from pix
    if (width < pixW + dotW) {
      // don't draw
      pixW = 0;
    }
    else
      pixDrawn = false;
  }

  w = pixW + fontMetrics().width(name);

  // if we have unused space from line above:
  // use it if whole name does fit in last line...
  if (unused>0) {
    //qDebug("DrawString: %s, '%s': Unused %d, StrW %d, Width %d",
    //       item->name().ascii(), name.ascii(), unused, w, width);
    if ((unused <= width) && (w <= unused)) {
      y = bottom ? (y+h) : (y-h);
      lines++;
    }
  }

  if (lines<1) return 0;

  p->setPen( (qGray(item->color().rgb())>100) ? black : white);
  if (rotate) {
    //p->translate(r.x()+2, r.y()+r.height());
    p->translate(r.x(), r.y()+r.height()-2);
    p->rotate(270);
  }
  else
    p->translate(r.x()+2, r.y());

  // loop over name parts to break up with small width.
  // every char category change is supposed a possible break,
  // with the exception Uppercase=>Lowercase.
  // It's good enough for numbers, Symbols...
  while (lines>0) {

    if (w>width && lines>1) {
      int lastBreakPos = name.length(), lastWidth = w;
      int len = name.length();
      QChar::Category caOld, ca;

      // start with comparing categories of last 2 chars...
      caOld = name[len-1].category();
      while (len>2) {
        len--;
        ca = name[len-1].category();
        if (ca != caOld) {
          // "Aa" has no break between...
          if (ca == QChar::Letter_Uppercase &&
              caOld == QChar::Letter_Lowercase) {
            caOld = ca;
            continue;
          }
          caOld = ca;
          lastBreakPos = len;
          w = pixW + fontMetrics().width(name, len);
          lastWidth = w;
          if (w <= width) break;
        }
      }
      w = lastWidth;
      remaining = name.mid(lastBreakPos);
      // remove space on break point
      if (name[lastBreakPos-1].category() == QChar::Separator_Space)
        name = name.left(lastBreakPos-1);
      else
        name = name.left(lastBreakPos);
    }
    else
      remaining = QString::null;

    /* truncate and add ... if needed */
    if (w>width) {
      int len = name.length();
      w += dotW;
      while (len>2 && (w > width)) {
        len--;
        w = pixW + fontMetrics().width(name, len) + dotW;
      }
      // stop drawing: we can't draw 2 chars + "..."
      if (w>width) break;

      name = name.left(len) + "...";
    }

    if (!pixDrawn) p->drawPixmap(0, y, pix);

    p->drawText(pixW, y, width-pixW, h,
                right ? Qt::AlignRight :Qt::AlignLeft, name);
    y = bottom ? (y-h) : (y+h);
    lines--;

    if (remaining.isEmpty()) break;
    name = remaining;
    w = pixW + fontMetrics().width(name);
  }

  if (!bottom) {
    if (rotate)
      r.setRect(r.x()+y, r.y(), r.width()-y, r.height());
    else
      r.setRect(r.x(), r.y()+y, r.width(), r.height()-y);
  }
  else {
    if (rotate)
      r.setRect(r.x(), r.y(), y+h, r.height());
    else
      r.setRect(r.x(), r.y(), r.width(), y+h);
  }
  p->resetXForm();

  if (w>width) w = width;
  return width-w;
}


/**
 * Draw TreeMapItems recursive, starting from item
 */
void TreeMapWidget::drawItems(QPainter* p,
                                 TreeMapItem* item)
{
  drawItem(p, item);

  QRect origRect = item->itemRect();
  int bw = item->borderWidth();
  QRect r = QRect(origRect.x()+bw, origRect.y()+bw,
                  origRect.width()-2*bw, origRect.height()-2*bw);

  TreeMapItemList* list = item->childs();
  TreeMapItem* i;

  bool stopDrawing = false;

  // only subdivide if there are childs
  if (!list || list->count()==0)
    stopDrawing = true;

  // only subdivide if there is enough space
  if (!stopDrawing && (r.width()<=0 || r.height()<=0))
    stopDrawing = true;

  // stop drawing if maximum depth is reached
  if (!stopDrawing &&
      (_maxDrawingDepth>=0 && item->depth()>=_maxDrawingDepth))
    stopDrawing = true;

  // stop drawing if stopAtText is reached
  if (!stopDrawing)
    for (int no=0;no<(int)_attr.size();no++) {
      QString stopAt = textStop(no);
      if (!stopAt.isEmpty() && (item->text(no) == stopAt)) {
        stopDrawing = true;
        break;
      }
    }

  // stop drawing if minimal area size is reached
  if (!stopDrawing &&
      (_minimalArea > 0) &&
      (r.width() * r.height() < _minimalArea)) stopDrawing = true;

  // not used at the moment (for bottom text)
  item->setFreeRect3(QRect());

  if (stopDrawing) {
    if (list) {
      // invalidate rects
      for (i=list->first();i;i=list->next())
        i->setItemRect(QRect());
    }
    // tooltip apears on whole item rect
    item->setFreeRect1(item->itemRect());
    item->setFreeRect2(QRect()); // not needed

    // if we have space...
    int unused[4] = {0, 0, 0, 0};
    int no;
    for (no=0;no<(int)_attr.size();no++) {
      if (!textVisible(no)) continue;
      switch(position(no)) {
      case TopLeft:
        unused[0] = drawString(p, item, no, r, unused[1], false, false);
        unused[1] = 0;
        break;
      case TopRight:
        unused[1] = drawString(p, item, no, r, unused[0], true, false);
        unused[0] = 0;
        break;
      case BottomLeft:
        unused[2] = drawString(p, item, no, r, unused[3], false, true);
        unused[3] = 0;
        break;
      case BottomRight:
        unused[3] = drawString(p, item, no, r, unused[2], true, true);
        unused[2] = 0;
        break;
      }
    }

    return;
  }

  unsigned user_sum, child_sum, self;

  // user supplied sum
  user_sum = item->sum();

  // own sum
  child_sum = 0;
  for (i=list->first();i;i=list->next())
    child_sum += i->value();

  QRect orig = r;
  int unused[4] = {0, 0, 0, 0};
  int no;
  for (no=0;no<(int)_attr.size();no++) {
    if (!textVisible(no)) continue;
    if (!textForced(no)) continue;
    switch(position(no)) {
    case TopLeft:
      unused[0] = drawString(p, item, no, r, unused[1], false, false);
      unused[1] = 0;
      break;
    case TopRight:
      unused[1] = drawString(p, item, no, r, unused[0], true, false);
      unused[0] = 0;
      break;
    case BottomLeft:
      unused[2] = drawString(p, item, no, r, unused[3], false, true);
      unused[3] = 0;
      break;
    case BottomRight:
      unused[3] = drawString(p, item, no, r, unused[2], true, true);
      unused[2] = 0;
      break;
    }
  }

  // set freeRect1 for tooltip
  if (orig.x() == r.x()) {
    // Strings on top
    item->setFreeRect1(QRect(orig.x(), orig.y(),
                             orig.width(), orig.height()-r.height()));
  }
  else {
    // Strings on the left
    item->setFreeRect1(QRect(orig.x(), orig.y(),
                             orig.width()-r.width(), orig.height()));
  }

  if (user_sum == 0) {
    // user didn't supply any sum
    user_sum = child_sum;
    self = 0;
  }
  else {
    self = user_sum - child_sum;

    if (user_sum < child_sum) {
      //qDebug("TreeMWidget %s: User sum %d < Child Items sum %d",
      //       item->path().ascii(), user_sum, child_sum);

      // invalid user supplied sum: ignore and use calculate sum
      user_sum = child_sum;
      self = 0;
    }
    else {
      // Try to put the border waste in self
      // percent of wasted space on border...
      float borderArea = origRect.width() * origRect.height();
      borderArea = (borderArea - r.width()*r.height())/borderArea;
      unsigned borderValue = (unsigned)(borderArea * user_sum);

      if (borderValue > self) {
        if (_skipIncorrectBorder) {
          r = origRect;
          // should add my self to nested self and set my self =0
        }
        else
          self = 0;
      }
      else
        self -= borderValue;

      user_sum = child_sum + self;
    }
  }

  bool goBack;
  if (item->sorting(&goBack) == -1) {
    // noSorting
    goBack = false;
  }

  TreeMapItemIterator it(*list);
  if (goBack) it.toLast();

  if (item->splitMode() == Columns) {
    int len = list->count();

    while (len>0 && user_sum>0) {
      TreeMapItemIterator first = it;
      unsigned valSum = 0;
      int lenLeft = len;
      int columns = (int)(sqrt((double)len * r.width()/r.height())+.5);
      if (columns==0) columns = 1; //should never be needed

      while (lenLeft>0 && ((double)valSum*(len-lenLeft) <
                           (double)len*user_sum/columns/columns)) {
        valSum += it.current()->value();
        if (goBack) --it; else ++it;
        lenLeft--;
      }

      // we always split horizontally
      int nextPos = (int)((double)r.width() * valSum / user_sum);
      QRect firstRect = QRect(r.x(), r.y(), nextPos, r.height());
      drawItemArray(p, item, firstRect,
                    valSum, first, len-lenLeft, goBack);
      r.setRect(r.x()+nextPos, r.y(), r.width()-nextPos, r.height());

      user_sum -= valSum;
      len = lenLeft;
    }
  }
  else if (item->splitMode() == Rows) {
    int len = list->count();

    while (len>0 && user_sum>0) {
      TreeMapItemIterator first = it;
      unsigned valSum = 0;
      int lenLeft = len;
      int rows = (int)(sqrt((double)len * r.height()/r.width())+.5);
      if (rows==0) rows = 1; //should never be needed

      while (lenLeft>0 && ((double)valSum*(len-lenLeft) <
                           (double)len*user_sum/rows/rows)) {
        valSum += it.current()->value();
        if (goBack) --it; else ++it;
        lenLeft--;
      }

      // we always split horizontally
      int nextPos = (int)((double)r.height() * valSum / user_sum);
      QRect firstRect = QRect(r.x(), r.y(), r.width(), nextPos);
      drawItemArray(p, item, firstRect,
                    valSum, first, len-lenLeft, goBack);
      r.setRect(r.x(), r.y()+nextPos, r.width(), r.height()-nextPos);

      user_sum -= valSum;
      len = lenLeft;
    }
  }
  else
    drawItemArray(p, item, r, user_sum, it, list->count(), goBack);


  // set freeRect (not occupied by childs) for tooltip
  r.setRect(r.x()-1, r.y()-1, r.width()+2, r.height()+2);
  item->setFreeRect2(r);

  for (no=0;no<(int)_attr.size();no++) {
    if (!textVisible(no)) continue;
    if (textForced(no)) continue;
    switch(position(no)) {
    case TopLeft:
      unused[0] = drawString(p, item, no, r, unused[1], false, false);
      unused[1] = 0;
      break;
    case TopRight:
      unused[1] = drawString(p, item, no, r, unused[0], true, false);
      unused[0] = 0;
      break;
    case BottomLeft:
      unused[2] = drawString(p, item, no, r, unused[3], false, true);
      unused[3] = 0;
      break;
    case BottomRight:
      unused[3] = drawString(p, item, no, r, unused[2], true, true);
      unused[2] = 0;
      break;
    }
  }
#if 0
  if (!textForced(0)) {
    // if we have space...
    int unused = 0;
    if (textVisible(0))
      unused = drawString(p, item, 0, r, 0, false, false);
    if (textVisible(1))
      unused = drawString(p, item, 1, r, unused, true, false);
    if (textVisible(2))
      drawString(p, item, 2, r, textVisible(0) ? 0 : unused, false, false);
  }
#endif
}


void TreeMapWidget::drawItemArray(QPainter* p, TreeMapItem* item,
                                  QRect& r, unsigned user_sum,
                                  TreeMapItemIterator it, int len, bool goBack)
{
  if (user_sum == 0) return;

  static bool b2t = true;
  unsigned scale = 1;
  unsigned tmp_sum = user_sum;

  while (tmp_sum>10000) {
    scale *= 10;
    tmp_sum /= 10;
  }

  if (len>2 && (item->splitMode() == Bisection)) {
    TreeMapItemIterator first = it;
    unsigned valSum = 0;
    int lenLeft = len;
    //while (lenLeft>0 && valSum<user_sum/2) {
    while (lenLeft>len/2) {
      valSum += it.current()->value();
      if (goBack) --it; else ++it;
      lenLeft--;
    }

    // draw first half...
    if (r.width() > r.height()) {
      int halfPos = (int)((double)r.width() * valSum / user_sum);
      QRect firstRect = QRect(r.x(), r.y(), halfPos, r.height());
      drawItemArray(p, item, firstRect,
                    valSum, first, len-lenLeft, goBack);
      r.setRect(r.x()+halfPos, r.y(), r.width()-halfPos, r.height());
    }
    else {
      int halfPos = (int)((double)r.height() * valSum / user_sum);
      QRect firstRect = QRect(r.x(), r.y(), r.width(), halfPos);
      drawItemArray(p, item, firstRect,
                    valSum, first, len-lenLeft, goBack);
      r.setRect(r.x(), r.y()+halfPos, r.width(), r.height()-halfPos);
    }

    // second half
    drawItemArray(p, item, r, user_sum - valSum, it, lenLeft, goBack);
    return;
  }

  bool hor = horizontal(item,r);
  int scaleCorrection = 0;

  TreeMapItem* i;
  while (len>0) {
    i = it.current();
    if (user_sum <= 0) {
      i->setItemRect(QRect());
      if (goBack) --it; else ++it;
      len--;
      continue;
    }

    if (i->splitMode() == AlwaysBest)
      hor = r.width() > r.height();

    int lastPos = hor ? r.width() : r.height();
    unsigned int val = i->value();
    int scaledSum = (int)(user_sum/scale);
    int scaledVal = (int)(val/scale) + scaleCorrection;
    int nextPos = (scaledSum==0) ? 0: (lastPos * scaledVal / scaledSum);
    if (nextPos>lastPos) nextPos = lastPos;
    // the rounding error of nextPos calculated back to value error
    // will be added to val in the next round
    int scaledOrig = (lastPos==0) ? 0 : (nextPos * scaledSum / lastPos);
    scaleCorrection = scaledVal - scaledOrig;

    // don't draw very small rectangles:
    if (nextPos >= _visibleWidth) {
      if (hor)
        i->setItemRect(QRect(r.x(), r.y(), nextPos, r.height()));
      else {
        if (b2t)
          i->setItemRect(QRect(r.x(), r.bottom()-nextPos+1,
                               r.width(), nextPos));
        else
          i->setItemRect(QRect(r.x(), r.y(), r.width(), nextPos));
      }

      /*
      qDebug("Item %s: R %d/%d-%d/%d, val %d/%d",
             i->path().ascii(),
             i->itemRect().x(), i->itemRect().y(),
             i->itemRect().width(), i->itemRect().height(),
             val, sum2);
      */

      drawItems(p, i);
    }
    else {
      // we can give our unused space the next item...
      if (_reuseSpace)
        nextPos = 0;
      i->setItemRect(QRect());
    }

    // draw Separator
    if (_drawSeparators && (nextPos<lastPos)) {
      p->setPen(black);
      if (hor) {
        if (r.top()<=r.bottom())
          p->drawLine(r.x() + nextPos, r.top(), r.x() + nextPos, r.bottom());
      }
      else {
        if (r.left()<=r.right())
        p->drawLine(r.left(), r.y() + nextPos, r.right(), r.y() + nextPos);
      }
      nextPos++;
    }

    if (hor)
      r.setRect(r.x() + nextPos, r.y(), lastPos-nextPos, r.height());
    else {
      if (b2t)
        r.setRect(r.x(), r.y(), r.width(), lastPos-nextPos);
      else
        r.setRect(r.x(), r.y() + nextPos, r.width(), lastPos-nextPos);
    }

    user_sum -= val;
    if (goBack) --it; else ++it;
    len--;
  }
}



/*----------------------------------------------------------------
 * Popup menus for option setting
 */

void TreeMapWidget::splitActivated(int id)
{
  if      (id == _splitID)   setSplitMode(Bisection);
  else if (id == _splitID+1) setSplitMode(Columns);
  else if (id == _splitID+2) setSplitMode(Rows);
  else if (id == _splitID+3) setSplitMode(AlwaysBest);
  else if (id == _splitID+4) setSplitMode(Best);
  else if (id == _splitID+5) setSplitMode(VAlternate);
  else if (id == _splitID+6) setSplitMode(HAlternate);
  else if (id == _splitID+7) setSplitMode(TreeMapItem::Horizontal);
  else if (id == _splitID+8) setSplitMode(TreeMapItem::Vertical);
}


QPopupMenu* TreeMapWidget::splitDirectionMenu(int id)
{
  _splitID = id;

  QPopupMenu* popup = new QPopupMenu();
  popup->setCheckable(true);

  connect(popup, SIGNAL(activated(int)),
          this, SLOT(splitActivated(int)));

  popup->insertItem(i18n("Recursive Bisection"), id);
  popup->insertItem(i18n("Columns"),         id+1);
  popup->insertItem(i18n("Rows"),            id+2);
  popup->insertItem(i18n("Always Best"),     id+3);
  popup->insertItem(i18n("Best"),            id+4);
  popup->insertItem(i18n("Alternate (V)"),   id+5);
  popup->insertItem(i18n("Alternate (H)"),   id+6);
  popup->insertItem(i18n("Horizontal"),      id+7);
  popup->insertItem(i18n("Vertical"),        id+8);

  switch(splitMode()) {
    case Bisection:  popup->setItemChecked(id,true); break;
    case Columns:    popup->setItemChecked(id+1,true); break;
    case Rows:       popup->setItemChecked(id+2,true); break;
    case AlwaysBest: popup->setItemChecked(id+3,true); break;
    case Best:       popup->setItemChecked(id+4,true); break;
    case VAlternate: popup->setItemChecked(id+5,true); break;
    case HAlternate: popup->setItemChecked(id+6,true); break;
    case TreeMapItem::Horizontal: popup->setItemChecked(id+7,true); break;
    case TreeMapItem::Vertical:   popup->setItemChecked(id+8,true); break;
    default: break;
  }
  return popup;
}

void TreeMapWidget::visualizationActivated(int id)
{
  if (id == _visID+2) setSkipIncorrectBorder(!skipIncorrectBorder());
  else if (id == _visID+3) setBorderWidth(0);
  else if (id == _visID+4) setBorderWidth(1);
  else if (id == _visID+5) setBorderWidth(2);
  else if (id == _visID+6) setBorderWidth(3);
  else if (id == _visID+10) setAllowRotation(!allowRotation());
  else if (id == _visID+11) setShadingEnabled(!isShadingEnabled());
  else if (id<_visID+19 || id>_visID+100) return;

  id -= 20+_visID;
  int textNo = id/10;
  if      ((id%10) == 1) setTextVisible(textNo, !textVisible(textNo));
  else if ((id%10) == 2) setTextForced(textNo, !textForced(textNo));
  else if ((id%10) == 3) setPosition(textNo, TreeMapItem::TopLeft);
  else if ((id%10) == 4) setPosition(textNo, TreeMapItem::TopRight);
  else if ((id%10) == 5) setPosition(textNo, TreeMapItem::BottomLeft);
  else if ((id%10) == 6) setPosition(textNo, TreeMapItem::BottomRight);
}

QPopupMenu* TreeMapWidget::visualizationMenu(int id)
{
  _visID = id;

  QPopupMenu* popup = new QPopupMenu();
  popup->setCheckable(true);

  QPopupMenu* bpopup = new QPopupMenu();
  bpopup->setCheckable(true);

  connect(popup, SIGNAL(activated(int)),
          this, SLOT(visualizationActivated(int)));
  connect(bpopup, SIGNAL(activated(int)),
          this, SLOT(visualizationActivated(int)));

  popup->insertItem(i18n("Nesting"),
                    splitDirectionMenu(id+100), id);

  popup->insertItem(i18n("Border"), bpopup, id+1);
  bpopup->insertItem(i18n("Correct Borders Only"), id+2);
  bpopup->insertSeparator();
  bpopup->insertItem(i18n("Border 0"), id+3);
  bpopup->insertItem(i18n("Border 1"), id+4);
  bpopup->insertItem(i18n("Border 2"), id+5);
  bpopup->insertItem(i18n("Border 3"), id+6);
  bpopup->setItemChecked(id+2, skipIncorrectBorder());
  bpopup->setItemChecked(id+3, borderWidth()==0);
  bpopup->setItemChecked(id+4, borderWidth()==1);
  bpopup->setItemChecked(id+5, borderWidth()==2);
  bpopup->setItemChecked(id+6, borderWidth()==3);

  popup->insertItem(i18n("Allow Rotation"), id+10);
  popup->setItemChecked(id+10,allowRotation());
  popup->insertItem(i18n("Shading"), id+11);
  popup->setItemChecked(id+11,isShadingEnabled());

  if (_attr.size() ==0) return popup;

  popup->insertSeparator();
  int textNo;
  QPopupMenu* tpopup;
  id += 20;
  for (textNo=0;textNo<(int)_attr.size();textNo++, id+=10) {
    tpopup = new QPopupMenu();
    tpopup->setCheckable(true);
    popup->insertItem(_attr[textNo].type, tpopup, id);
    tpopup->insertItem(i18n("Visible"), id+1);
    tpopup->insertItem(i18n("Take Space from Children"), id+2);
    tpopup->insertSeparator();
    tpopup->insertItem(i18n("Top Left"), id+3);
    tpopup->insertItem(i18n("Top Right"), id+4);
    tpopup->insertItem(i18n("Bottom Left"), id+5);
    tpopup->insertItem(i18n("Bottom Right"), id+6);

    tpopup->setItemChecked(id+1,_attr[textNo].visible);
    tpopup->setItemEnabled(id+2,_attr[textNo].visible);
    tpopup->setItemEnabled(id+3,_attr[textNo].visible);
    tpopup->setItemEnabled(id+4,_attr[textNo].visible);
    tpopup->setItemEnabled(id+5,_attr[textNo].visible);
    tpopup->setItemEnabled(id+6,_attr[textNo].visible);
    tpopup->setItemChecked(id+2,_attr[textNo].forced);
    tpopup->setItemChecked(id+3,_attr[textNo].pos == TopLeft);
    tpopup->setItemChecked(id+4,_attr[textNo].pos == TopRight);
    tpopup->setItemChecked(id+5,_attr[textNo].pos == BottomLeft);
    tpopup->setItemChecked(id+6,_attr[textNo].pos == BottomRight);

    connect(tpopup, SIGNAL(activated(int)),
            this, SLOT(visualizationActivated(int)));
  }

  return popup;
}

void TreeMapWidget::selectionActivated(int id)
{
  TreeMapItem* i = _menuItem;
  id -= _selectionID;
  while (id>0 && i) {
    i=i->parent();
    id--;
  }
  if (i)
    setSelected(i, true);
}

QPopupMenu* TreeMapWidget::selectionMenu(int id, TreeMapItem* i)
{
  if (!i) return 0;

  _selectionID = id;
  _menuItem = i;
  QPopupMenu* popup = new QPopupMenu();

  connect(popup, SIGNAL(activated(int)),
          this, SLOT(selectionActivated(int)));

  while (i) {
    QString name = i->text(0);
    if (name.isEmpty()) break;
    popup->insertItem(i->text(0), id++);
    i = i->parent();
  }

  return popup;
}

void TreeMapWidget::textStopActivated(int id)
{
  if (id == _textStopID) setTextStop(0, QString::null);
  else {
    TreeMapItem* i = _menuItem;
    id -= _textStopID+1;
    while (id>0 && i) {
      i=i->parent();
      id--;
    }
    if (i)
      setTextStop(0, i->text(0));
  }
}

QPopupMenu* TreeMapWidget::textStopMenu(int id, TreeMapItem* i)
{
  _textStopID = id;
  QPopupMenu* popup = new QPopupMenu();
  connect(popup, SIGNAL(activated(int)),
          this, SLOT(textStopActivated(int)));

  popup->insertItem(i18n("No %1 Limit").arg(textType(0)), id);
  popup->setItemChecked(id, textStop(0).isEmpty());
  _menuItem = i;
  bool foundTextStop = false;
  if (i) {
    popup->insertSeparator();

    while (i) {
      id++;
      QString name = i->text(0);
      if (name.isEmpty()) break;
      popup->insertItem(i->text(0), id);
      if (textStop(0) == i->text(0)) {
        popup->setItemChecked(id, true);
        foundTextStop = true;
      }
      i = i->parent();
    }
  }

  if (!foundTextStop && !textStop(0).isEmpty()) {
    popup->insertSeparator();
    popup->insertItem(textStop(0), id+1);
    popup->setItemChecked(id+1, true);
  }

  return popup;
}

void TreeMapWidget::areaStopActivated(int id)
{
  if (id == _areaStopID) setMinimalArea(-1);
  else if (id == _areaStopID+1) {
    int area = _menuItem ? (_menuItem->width() * _menuItem->height()) : -1;
    setMinimalArea(area);
  }
  else if (id == _areaStopID+2) setMinimalArea(100);
  else if (id == _areaStopID+3) setMinimalArea(400);
  else if (id == _areaStopID+4) setMinimalArea(1000);
  else if (id == _areaStopID+5) setMinimalArea(minimalArea()*2);
  else if (id == _areaStopID+6) setMinimalArea(minimalArea()/2);
}

QPopupMenu* TreeMapWidget::areaStopMenu(int id, TreeMapItem* i)
{
  _areaStopID = id;
  _menuItem = i;

  QPopupMenu* popup = new QPopupMenu();
  connect(popup, SIGNAL(activated(int)),
          this, SLOT(areaStopActivated(int)));

  bool foundArea = false;

  popup->insertItem(i18n("No Area Limit"), id);
  popup->setItemChecked(id, minimalArea() == -1);

  if (i) {
    int area = i->width() * i->height();
    popup->insertSeparator();
    popup->insertItem(i18n("Area of '%1' (%2)")
                       .arg(i->text(0)).arg(area), id+1);
    if (area == minimalArea()) {
      popup->setItemChecked(id+1, true);
      foundArea = true;
    }
  }

  popup->insertSeparator();
  int area = 100, count;
  for (count=0;count<3;count++) {
    popup->insertItem(i18n("%1 Pixels").arg(area), id+2+count);
    if (area == minimalArea()) {
      popup->setItemChecked(id+2+count, true);
      foundArea = true;
    }
    area = (area==100) ? 400 : (area==400) ? 1000 : 4000;
  }

  if (minimalArea()>0) {
    popup->insertSeparator();
    if (!foundArea) {
      popup->insertItem(i18n("%1 Pixels").arg(minimalArea()), id+10);
      popup->setItemChecked(id+10, true);
    }

    popup->insertItem(i18n("Double Area Limit (to %1)")
                       .arg(minimalArea()*2), id+5);
    popup->insertItem(i18n("Half Area Limit (to %1)")
                       .arg(minimalArea()/2), id+6);
  }

  return popup;
}


void TreeMapWidget::depthStopActivated(int id)
{
  if (id == _depthStopID) setMaxDrawingDepth(-1);
  else if (id == _depthStopID+1) {
    int d = _menuItem ? _menuItem->depth() : -1;
    setMaxDrawingDepth(d);
  }
  else if (id == _depthStopID+2) setMaxDrawingDepth(maxDrawingDepth()-1);
  else if (id == _depthStopID+3) setMaxDrawingDepth(maxDrawingDepth()+1);
}

QPopupMenu* TreeMapWidget::depthStopMenu(int id, TreeMapItem* i)
{
  _depthStopID = id;
  _menuItem = i;

  QPopupMenu* popup = new QPopupMenu();
  connect(popup, SIGNAL(activated(int)),
          this, SLOT(depthStopActivated(int)));

  bool foundDepth = false;

  popup->insertItem(i18n("No Depth Limit"), id);
  popup->setItemChecked(id, maxDrawingDepth() == -1);

  if (i) {
    int d = i->depth();
    popup->insertSeparator();
    popup->insertItem(i18n("Depth of '%1' (%2)")
                       .arg(i->text(0)).arg(d), id+1);
    if (d == maxDrawingDepth()) {
      popup->setItemChecked(id+1, true);
      foundDepth = true;
    }
  }

  if (maxDrawingDepth()>1) {
    popup->insertSeparator();
    if (!foundDepth) {
      popup->insertItem(i18n("Depth %1").arg(maxDrawingDepth()), id+10);
      popup->setItemChecked(id+10, true);
    }

    popup->insertItem(i18n("Decrement (to %1)")
                       .arg(maxDrawingDepth()-1), id+2);
    popup->insertItem(i18n("Increment (to %1)")
                       .arg(maxDrawingDepth()+1), id+3);
  }

  return popup;
}



/*----------------------------------------------------------------
 * Option saving/restoring
 */

void TreeMapWidget::saveOptions(KConfigGroup* config, QString prefix)
{
  config->writeEntry(prefix+"Nesting", splitModeString());
  config->writeEntry(prefix+"AllowRotation", allowRotation());
  config->writeEntry(prefix+"ShadingEnabled", isShadingEnabled());
  config->writeEntry(prefix+"OnlyCorrectBorder", skipIncorrectBorder());
  config->writeEntry(prefix+"BorderWidth", borderWidth());
  config->writeEntry(prefix+"MaxDepth", maxDrawingDepth());
  config->writeEntry(prefix+"MinimalArea", minimalArea());

  int textNo, textNoCount = _attr.size();
  config->writeEntry(prefix+"TextCount", textNoCount);
  for (textNo=0;textNo<textNoCount;textNo++) {
    config->writeEntry(QString(prefix+"TextVisible%1").arg(textNo),
                       _attr[textNo].visible);
    config->writeEntry(QString(prefix+"TextForced%1").arg(textNo),
                       _attr[textNo].forced);
    config->writeEntry(QString(prefix+"TextStop%1").arg(textNo),
                       _attr[textNo].stop);
    config->writeEntry(QString(prefix+"Position%1").arg(textNo),
                       positionString(textNo));
  }
}


void TreeMapWidget::restoreOptions(KConfigGroup* config, QString prefix)
{
  bool enabled;
  int num;
  QString str;

  str = config->readEntry(prefix+"Nesting");
  if (!str.isEmpty()) setSplitMode(str);

  if (config->hasKey(prefix+"AllowRotation")) {
    enabled = config->readBoolEntry(prefix+"AllowRotation", true);
    setAllowRotation(enabled);
  }

  if (config->hasKey(prefix+"ShadingEnabled")) {
    enabled = config->readBoolEntry(prefix+"ShadingEnabled", true);
    setShadingEnabled(enabled);
  }

  if (config->hasKey(prefix+"OnlyCorrectBorder")) {
    enabled = config->readBoolEntry(prefix+"OnlyCorrectBorder", false);
    setSkipIncorrectBorder(enabled);
  }

  num = config->readNumEntry(prefix+"BorderWidth", -2);
  if (num!=-2) setBorderWidth(num);

  num = config->readNumEntry(prefix+"MaxDepth", -2);
  if (num!=-2) setMaxDrawingDepth(num);

  num = config->readNumEntry(prefix+"MinimalArea", -2);
  if (num!=-2) setMinimalArea(num);

  num = config->readNumEntry(prefix+"TextCount", -2);
  if (num<=0 || num>MAX_TREEMAP_TEXT) return;

  int textNo;
  for (textNo=0;textNo<num;textNo++) {
    str = QString(prefix+"TextVisible%1").arg(textNo);
    if (config->hasKey(str))
      setTextVisible(textNo, config->readBoolEntry(str));

    str = QString(prefix+"TextForced%1").arg(textNo);
    if (config->hasKey(str))
      setTextForced(textNo, config->readBoolEntry(str));

    str = config->readEntry(QString(prefix+"TextStop%1").arg(textNo));
    setTextStop(textNo, str);

    str = config->readEntry(QString(prefix+"Position%1").arg(textNo));
    if (!str.isEmpty()) setPosition(textNo, str);
  }
}
