/**
 * A Widget for visualizing hierarchical metrics as areas.
 * The API is similar to QListView.
 *
 * JW, 2002
 */

#ifndef TREEMAP_H
#define TREEMAP_H

#include <qstring.h>
#include <qwidget.h>
#include <qpixmap.h>
#include <qptrlist.h>
#include <qvaluevector.h>
#include <qcolor.h>
#include <qstringlist.h>

class QPopupMenu;
class TreeMapTip;
class TreeMapWidget;
class TreeMapItem;
class TreeMapItemList;
class QString;

class KConfigGroup;

class TreeMapItemList: public QPtrList<TreeMapItem>
{
protected:
  int compareItems ( Item item1, Item item2 );
};

typedef QPtrListIterator<TreeMapItem> TreeMapItemIterator;

/**
 * Base class of items in TreeMap.
 *
 * This class supports an arbitrary number of text() strings
 * positioned counterclock-wise starting at TopLeft. Each item
 * has a static value(), sum() and color(). The sorting(),
 * splitMode() and borderWidth() is taken from a TreeMap.
 *
 * If you want more flexibility, reimplement TreeMapItem and
 * override the corresponding methods. For dynamic creation of child
 * items on demand, reimplement childs().
 */
class TreeMapItem
{
public:

  /**
   * Split direction for nested areas:
   *  AlwaysBest: Choose split direction for every subitem according to
   *              longest side of rectangle left for drawing
   *  Best:       Choose split direction for all subitems of an area
   *              depending on longest side
   *  HAlternate: Horizontal at top;  alternate direction on depth step
   *  VAlternate: Vertical at top; alternate direction on depth step
   *  Horizontal: Always horizontal split direction
   *  Vertical:   Always vertical split direction
   */
  enum SplitMode { Bisection, Columns, Rows,
                   AlwaysBest, Best,
                   HAlternate, VAlternate,
                   Horizontal, Vertical };

  /**
   * Position of text drawn into a TreeMapItem.
   *
   * The specified position assumes no item rotation.
   * If there is more than one text for one position, it is put
   * nearer to the center of the item.
   */
  enum Position { TopLeft, TopRight, BottomLeft, BottomRight};


  TreeMapItem(TreeMapItem* parent = 0, unsigned int value = 1 );
  TreeMapItem(TreeMapItem* parent, unsigned int value,
              QString text1, QString text2 = QString::null,
              QString text3 = QString::null, QString text4 = QString::null);
  virtual ~TreeMapItem();

  bool isChildOf(TreeMapItem*);

  TreeMapItem* commonParent(TreeMapItem* item);

  // force a redraw of this item
  void redraw();

  // force new child generation & refresh
  void refresh();

  // call in a reimplemented items() method to check if already called
  bool initialized();

  /**
   * Adds an item to a parent.
   * When no sorting is used, the item is appended (drawn at bottom).
   * This is only needed if the parent was not already specified in the
   * construction of the item.
   */
  void addItem(TreeMapItem*);

  /**
   * Returns a list of text strings of specified text number,
   * from root up to this item.
   */
  QStringList path(int);

  /**
   * Depth of this item. This is the distance to root.
   */
  int depth();

  /**
   * Parent Item
   */
  TreeMapItem* parent() { return _parent; }

  /**
   * Temporary rectangle used for drawing this item the last time.
   */
  const QRect& itemRect() { return _rect; }
  int width() { return _rect.width(); }
  int height() { return _rect.height(); }

  // internal, used for drawing
  const QRect& freeRect1() { return _free1; }
  const QRect& freeRect2() { return _free2; }
  const QRect& freeRect3() { return _free3; }

  /**
   * Temporary child item index of the child that was current() recently.
   */
  int index() { return _index; }

  /**
   * TreeMap widget this item is put in.
   */
  TreeMapWidget* widget() { return _widget; }

  void setParent(TreeMapItem* p);
  void setWidget(TreeMapWidget* w) { _widget = w; }
  void setText(int, QString);
  void setPixmap(int, QPixmap);
  void setSum(unsigned int s) { _sum = s; }
  void setValue(unsigned int s) { _value = s; }
  void setColor(QColor c) { _color = c; }

  // internal
  void setItemRect(const QRect& r) { _rect = r; }
  void setFreeRect1(const QRect& r) { _free1 = r; }
  void setFreeRect2(const QRect& r) { _free2 = r; }
  void setFreeRect3(const QRect& r) { _free3 = r; }
  void setIndex(int i) { _index = i; }

  virtual unsigned int sum();
  virtual unsigned int value();

  virtual QString text(int);
  virtual QPixmap pixmap(int);
  virtual Position position(int);
  virtual int textCount() { return _text.size(); }

  virtual int borderWidth();

  /**
   * Returns the text number after that sorting is done or
   * -1 for no sorting, -2 for value() sorting (default).
   * If ascending != 0, a bool value is written at that location
   * to indicate if sorting should be ascending.
   */
  virtual int sorting(bool* ascending);

  virtual QColor color();
  virtual SplitMode splitMode();
  virtual int rtti();
  virtual TreeMapItemList* childs();

protected:
  TreeMapItemList* _childs;
  unsigned int _sum, _value;
  QValueVector<QString> _text;
  QValueVector<QPixmap> _pix;
  QColor _color;

private:
  TreeMapWidget* _widget;
  TreeMapItem* _parent;

  // temporary layout
  QRect _rect, _free1, _free2, _free3;
  int _depth;

  // temporary self value (when using level skipping)
  unsigned int _unused_self;

  // index of last active subitem
  int _index;
};


/**
 * Class for visualisation of a metric of hierarchically
 * nested items as 2D areas.
 */
class TreeMapWidget: public QWidget,  public TreeMapItem
{
  Q_OBJECT

public:

  /**
   * Same as in QListBox/QListView
   */
  enum SelectionMode { Single, Multi, Extended, NoSelection };

  TreeMapWidget(QWidget* parent=0, const char* name=0);
  ~TreeMapWidget();

  /**
   * Returns the area item at position x/y, independent from any
   * maxSelectDepth setting.
   */
  TreeMapItem* item(int x, int y);

  /**
   * Returns the nearest item with a visible area; this
   * can be the given item itself.
   */
  TreeMapItem* visibleItem(TreeMapItem*);

  /**
   * Returns the item possible for selection. this returnds the
   * given item itself or a parent thereof,
   * depending on setting of maxSelectDepth().
   */
  TreeMapItem* possibleSelection(TreeMapItem*);

  /**
   * Selects or unselects an item.
   * In multiselection mode, the constrain that a selected item
   * has no selected childs or parents stays true.
   */
  void setSelected(TreeMapItem*, bool selected = true);

  /**
   * Clear selection of all selected items which are childs of
   * parent. When parent == 0, clears whole selection
   */
  void clearSelection(TreeMapItem* parent = 0);

  /**
   * Selects or unselects items in a range.
   * This is needed internally for Shift-Click in Extented mode.
   * Range means for a hierarchical widget:
   * - select/unselect i1 and i2 according selected
   * - search common parent of i1 and i2, and select/unselect the
   *   range of direct children between but excluding the child
   *   leading to i1 and the child leading to i2.
   */
  void setRangeSelection(TreeMapItem* i1,
                         TreeMapItem* i2, bool selected);

  /**
   * Sets the current item.
   * The current item is mainly used for keyboard navigation.
   */
  void setCurrent(TreeMapItem*, bool kbd=false);

  /**
   * Set the maximal depth a selected item can have.
   * If you try to select a item with higher depth, the ancestor holding
   * this condition is used.
   *
   * See also possibleSelection().
   */
  void setMaxSelectDepth(int d) { _maxSelectDepth = d; }


  void setSelectionMode(SelectionMode m) { _selectionMode = m; }

  /**
   * for setting/getting split direction
   */
  void setSplitMode(SplitMode m);
  SplitMode splitMode();
  // returns true if string was recognized
  bool setSplitMode(QString);
  QString splitModeString();


  /*
   * Shading of rectangles enabled ?
   */
  void setShadingEnabled(bool s);
  bool isShadingEnabled() { return _shading; }


  /**
   * Items usually have a size proportional to their value().
   * With <width>, you can give the minimum width
   * of the resulting rectangle to still be drawn.
   * For space not used because of to small items, you can specify
   * with <reuseSpace> if the background should shine through or
   * the space will be used to enlarge the next item to be drawn
   * at this level.
   */
  void setVisibleWidth(int width, bool reuseSpace = false);

  /**
   * If a childs value() is almost the parents sum(),
   * it can happen that the border to be drawn for visibilty of
   * nesting relations takes to much space, and the
   * parent/child size relation can not be mapped to a correct
   * area size relation.
   *
   * Either
   * (1) Ignore the incorrect drawing, or
   * (2) Skip drawing of the parent level alltogether.
   */
  void setSkipIncorrectBorder(bool enable = true);
  bool skipIncorrectBorder() { return _skipIncorrectBorder; }

  /**
   * Maximal nesting depth
   */
  void setMaxDrawingDepth(int d);
  int maxDrawingDepth() { return _maxDrawingDepth; }

  /**
   * Minimal area for rectangles to draw
   */
  void setMinimalArea(int area);
  int minimalArea() { return _minimalArea; }

  /* defaults for text attributes */
  QString defaultTextType(int);
  QString defaultTextStop(int);
  bool    defaultTextVisible(int);
  bool    defaultTextForced(int);
  Position defaultPosition(int);

  /**
   * Set the type name of the text.
   * This is important for the visualization menu generated
   * with visualizationMenu()
   */
  void setTextType(int, QString);
  QString textType(int);

  /**
   * Stop drawing at item with name
   */
  void setTextStop(int textNo, QString);
  QString textStop(int textNo);

  /**
   * Should the text with number textNo be visible?
   * This is only done if remaining space is enough to allow for
   * proportional size constrains.
   */
  void setTextVisible(int, bool);
  bool textVisible(int);

  /**
   * Should the drawing of the name into the rectangle be forced?
   * This enables drawing of the name before drawing subitems, and
   * thus destroys proportional constrains.
   */
  void setTextForced(int, bool);
  bool textForced(int);

  /**
   * Set the text position in the area. See TreeMapItem::Position
   */
  void setPosition(int, TreeMapItem::Position);
  TreeMapItem::Position position(int);
  void setPosition(int, QString);
  QString positionString(int);

  /**
   * Do we allow the texts to be rotated by 90 degrees for better fitting?
   */
  void setAllowRotation(bool);
  bool allowRotation() { return _allowRotation; }

  void setSorting(int textNo, bool ascending = true);
  int sorting(bool* ascending);

  void setBorderWidth(int w) { _borderWidth = w; }
  int borderWidth() { return _borderWidth; }

  /**
   * Save/restore options.
   */
  void saveOptions(KConfigGroup*, QString prefix = QString::null);
  void restoreOptions(KConfigGroup*, QString prefix = QString::null);

  /**
   * Popups for the context menu.
   * 1st int is the menu id where to start for the items
   */
  QPopupMenu* splitDirectionMenu(int);
  QPopupMenu* selectionMenu(int, TreeMapItem*);
  QPopupMenu* textStopMenu(int, TreeMapItem*);
  QPopupMenu* areaStopMenu(int, TreeMapItem*);
  QPopupMenu* depthStopMenu(int, TreeMapItem*);
  QPopupMenu* visualizationMenu(int);


  TreeMapWidget* widget() { return this; }
  TreeMapItem* current() { return _current; }
  TreeMapItemList selection() { return _selection; }
  bool isSelected(TreeMapItem* i);
  int maxSelectDepth() { return _maxSelectDepth; }
  SelectionMode selectionMode() { return _selectionMode; }

  /**
   * Return tooltip string to show for a item (can be rich text)
   * Default implementation gives lines with "text0 (text1)" going to root.
   */
  virtual QString tipString(TreeMapItem* i);

  /**
   * Redraws an item with all childs.
   * This takes changed values(), sums(), colors() and text() into account.
   */
  void redraw(TreeMapItem*);
  void redraw() { redraw(this); }

  // internal
  void drawTreeMap();

  // used internally when items are destroyed
  void deletingItem(TreeMapItem*);

protected slots:
  void splitActivated(int);
  void selectionActivated(int);
  void textStopActivated(int);
  void areaStopActivated(int);
  void depthStopActivated(int);
  void visualizationActivated(int);

signals:
  void selectionChanged();
  void selectionChanged(TreeMapItem*);

  /**
   * This signal is emitted if the current item changes.
   * If the change is done because of keyboard navigation,
   * the <kbd> is set to true
   */
  void currentChanged(TreeMapItem*, bool keyboard);

  void doubleClicked(TreeMapItem*);
  void rightButtonPressed(TreeMapItem*, const QPoint &);
  void contextMenuRequested(TreeMapItem*, const QPoint &);

protected:
  void mousePressEvent( QMouseEvent * );
  void contextMenuEvent( QContextMenuEvent * );
  void mouseReleaseEvent( QMouseEvent * );
  void mouseMoveEvent( QMouseEvent * );
  void mouseDoubleClickEvent( QMouseEvent * );
  void keyPressEvent( QKeyEvent* );
  void paintEvent( QPaintEvent * );
  void resizeEvent( QResizeEvent * );
  void showEvent( QShowEvent * );

private:
  // returns true if selection changed
  bool setTmpSelected(TreeMapItem*, bool selected = true);
  bool setTmpRangeSelection(TreeMapItem* i1,
                            TreeMapItem* i2, bool selected);
  bool isTmpSelected(TreeMapItem* i);

  void drawItem(QPainter* p, TreeMapItem*);
  int drawString(QPainter* p, TreeMapItem* item, int textNo, QRect& r,
                 int lastUnused, bool right, bool bottom);
  void drawItems(QPainter* p, TreeMapItem*);
  bool horizontal(TreeMapItem* i, const QRect& r);
  void drawItemArray(QPainter* p, TreeMapItem*, QRect& r, unsigned,
                     TreeMapItemIterator it, int len, bool);
  bool resizeAttr(int);

  TreeMapItem *_current, *_pressed, *_lastOver, *_oldCurrent;
  TreeMapTip* _tip;
  int _maxSelectDepth, _maxDrawingDepth;

  // attributes for text, per textNo
  struct TextAttr {
    QString type, stop;
    bool visible, forced;
    Position pos;
  };
  QValueVector<TextAttr> _attr;

  SelectionMode _selectionMode;
  SplitMode _splitMode;
  int _visibleWidth, _stopArea, _sortTextNo, _minimalArea, _borderWidth;
  bool _reuseSpace, _skipIncorrectBorder, _drawSeparators, _shading;
  bool _allowRotation, _sortAscending;
  TreeMapItem * _needsRefresh;
  TreeMapItemList _selection;

  // for the context menus: start IDs
  int _splitID, _selectionID, _visID;
  int _textStopID, _areaStopID, _depthStopID;
  TreeMapItem* _menuItem;

  // temporary selection while dragging, used for drawing
  // most of the time, _selection == _tmpSelection
  TreeMapItemList _tmpSelection;
  bool _inShiftDrag, _inControlDrag;

  QPixmap _pixmap;
};

#endif
